import json
from collections import OrderedDict


class ParameterSetMeta(type):

    def __new__(mcs, cls_name, base_classes, dct):
        for base in base_classes:
            if hasattr(base, "__metaclass__") and base.__metaclass__ == ParameterSetMeta:
                items = []
                for (name, attr) in dct.items():
                    if isinstance(attr, ConfigItem):
                        attr._name = name
                        items.append(attr)
                items.sort(key=lambda item: item._order)
                dct["_items"] = OrderedDict((item._name, item) for item in items)
                title = dct.get("title", "NO TITLE GIVEN")
                details = dct.get("details")
                if details:
                    dct["__doc__"] = title + "\n" + details
                else:
                    dct["__doc__"] = title

        return type.__new__(mcs, cls_name, base_classes, dct)

    def __init__(clz, name, base_classes, dct):
        clz._checks = {}
        for (name, attr) in dct.items():
            if name.startswith("check_"):
                field = name.lstrip("check_")
                if field not in dct:
                    raise Exception("field for checker %s unknwon" % name)
                # getattr(clz, field).check = lambda value, attr=attr: attr(value)
                clz._checks[field] = lambda value, attr=attr: attr(value)
        super(ParameterSetMeta, clz).__init__(name, base_classes, dct)


class ConfigItem(object):

    def __init__(self, name=None):
        from . import _context_stack
        ctx = _context_stack.current_context()
        self._order = ctx.new_order()
        self._name = name
        # stack as tuple: no chance to get into trouble with references to
        # ConfigItem.current_group_stack:
        self._group_stack = tuple(ctx.current_group_stack)


class ParameterSet(ConfigItem):
    __metaclass__ = ParameterSetMeta

    def __init__(self, name=None, expert_settings=False):
        if name is None:
            name = self.__class__.__name__
        super(ParameterSet, self).__init__(name=name)
        self._expert_settings = expert_settings

    def print_(self, indent=0):
        for item in self._items.values():
            if isinstance(item, ParameterSet):
                txt = "{}: {}".format(item._name, item.__class__.__name__)
                print " " * (indent - 1), txt
                item.print_(indent + 4)
            else:
                txt = "{}: {} = {!r}".format(item._name, item.__class__.__name__, item.value)
                print " " * (indent - 1), txt

    def check(self, name, value, item):
        ok = False
        result = item.check_type(value)
        if isinstance(result, (int, long,  bool)):
            ok = result
        else:
            raise RuntimeError("%r must return bool value" % item.check)
        if not ok:
            raise ValueError("can not set %s to %r because type check fails" % (name, value))

        ok = False
        result = item.check_value(value)
        if isinstance(result, (int, long,  bool)):
            ok = result
        else:
            raise RuntimeError("%r must return bool value" % item.check)
        if not ok:
            raise ValueError("can not set %s to %r because value check fails" % (name, value))

    def __iter__(self):
        dct = {}
        for name, item in self._items.items():
            if isinstance(item, ParameterSet):
                yield name, dict(item)
            else:
                yield name, item.value

    def __str__(self):
        shift = "    "

        def collect(obj, lines, indent="", shift=shift):
            for item in obj._items.values():
                if isinstance(item, ParameterSet):
                    lines.append(indent + "%s: {" % item._name)
                    collect(item, lines, indent + shift)
                    lines.append(indent + "}")
                else:
                    lines.append(indent + "%s = %s" % (item._name, item.value))
        lines = []
        lines.append("{")
        collect(self, lines, indent=shift)
        lines.append("}")
        return "\n".join(lines)

    def __setattr__(self, name, value):
        item = self._items.get(name)

        # recurse for nested parameter set:
        if isinstance(value, ParameterSet):
            for sub_name, sub_value in value:
                setattr(item, sub_name, sub_value)
            return

        item = self._items.get(name)
        if isinstance(self, ParameterSet) and not name.startswith("_") and item is None:
            raise Exception("value %s not known" % name)

        if item is not None and value is not None:
            # value = item.setup_value(value)
            if item.target is not None:
                to, method = item.target
                to.value = method(value)
            item.value = value
            return
        ConfigItem.__setattr__(self, name, value)

    def update(self, other):
        for name, item in other._items.items():
            if isinstance(item, ParameterSet):
                getattr(self, name).update(item)
            else:
                setattr(self, name, item.value)

    def update_from_dict(self, dd):
        """ maybe partial update if dd does not contain all value/parameter set items we defined """
        for name, item in self._items.items():
            if isinstance(item, ParameterSet):
                item.update_from_dict(dd.get(name, {}))
            else:
                old_value = getattr(self, name)
                setattr(self, name, dd.get(name, old_value))

    def overwrite_from_dict(self, dd):
        """ full update, dd must conatin al value/parameter set items we defined """
        for name, item in self._items.items():
            if name not in dd:
                continue
            if isinstance(item, ParameterSet):
                item.overwrite_from_dict(dd[name])
            else:
                setattr(self, name, dd[name])

    def as_dict(self):
        return dict(self)   # uses self.__iter__

    def __getitem__(self, name):
        return getattr(self, name)

    def __getstate__(self):
        return (self.__dict__, self.as_dict())

    def __setstate__(self, (dd, values)):
        self.__dict__ = dd
        self.overwrite_from_dict(values)

    def dumps(self):
        return json.dumps(self.as_dict(), indent=4)   # indent for pretty printing

    def dump_json(self, path):
        with open(path, "w") as fp:
            fp.write(self.dumps())

    def update_from_json(self, str_):
        self.overwrite_from_dict(json.loads(str_))

    def update_from_json_file(self, path):
        with open(path, "rU") as fp:
            return self.update_from_json(fp.read())

    @classmethod
    def loads(clz, str_):
        obj = clz()
        obj.update_from_json(str_)
        return obj

    @classmethod
    def load_json(clz, path):
        obj = clz()
        obj.update_from_json_file(path)
        return obj

    def as_widget(self, use_tabs=False):
        from .gui_creation import create_widget
        return create_widget(self, use_tabs)

    def edit(self, use_tabs=False):
        from guidata import qapplication
        app = qapplication()
        from .gui_creation import create_dlg
        backup = self.as_dict()
        flag = create_dlg(self, use_tabs).edit()
        if flag == 0:  # abort self might be edited, so we use backup for restore
            self.overwrite_from_dict(backup)
        return flag



from .patches import apply_patches
apply_patches()

from .module_globals import _context_stack

from builder import ParameterSetBuilder

import sys

# DO NOT TOUCH THE FOLLOWING LINE:
if not getattr(sys, "frozen", False):
    import pkg_resources  # part of setuptools
    __version__ = tuple(map(int, pkg_resources.require(__name__)[0].version.split(".")))
else:
    __version__ = (0, 0, 0)



import contextlib


def propagate_change(from_, to, method=None):
    if method is None:
        method = lambda x: x
    from_.target = (to, method)


def propagate_default(from_, to, method=None):
    if method is None:
        method = lambda x: x
    to._value = from_.value


@contextlib.contextmanager
def ValueGroup(name, expert_settings=False):
    from . import _context_stack
    ctx = _context_stack.current_context()
    ctx.current_group_stack.append(name)
    group_stack = tuple(ctx.current_group_stack)
    if expert_settings:
        ctx.expert_groups.add(group_stack)
    yield
    del ctx.current_group_stack[-1]





class _Context(object):

    def __init__(self):
        self.counter = -1
        self.current_group_stack = []
        self.expert_groups = set()

    def new_order(self):
        self.counter += 1
        return self.counter


class _ContextStack(object):

    """module global management of contexts.

    I'm not sure if this is realy needed, but if one accidentaly imports a module within
    a context and the module inturn creates a new context this would lead to very hard to
    find errors.
    """

    def __init__(self):
        self.stack = []

    def install_new_context(self):
        self.ctx = _Context()
        self.stack.append(self.ctx)

    def current_context(self):
        return self.ctx

    def drop_context(self):
        del self.stack[-1]

# this is a private module global variable:
_context_stack = _ContextStack()



# encoding: utf-8

import guidata.py3compat
import guidata.utils

import guidata.dataset.dataitems as di
import guidata.dataset.qtitemwidgets as qtiw
import guidata.dataset.qtwidgets as qtw

from PyQt4.QtGui import QWidget
from PyQt4.QtCore import SIGNAL


def patch_bugs():
    """patch bugs in guidata 1.6.1 for handling of empty file/directory fields"""

    def to_text_string(obj, encoding=None, _orig_function=guidata.py3compat.to_text_string):
        if obj is None:
            return None
        return _orig_function(obj, encoding)

    guidata.py3compat.to_text_string = to_text_string

    def add_extension(item, value, _orig_function=guidata.utils.add_extension):
        if value is None:
            return None
        return _orig_function(item, value)

    guidata.utils.add_extension = add_extension

    def from_string(self, value):
        value = to_text_string(value)
        if value is None:
            return []
        if value.endswith("']"):
            value = eval(value)
        else:
            value = [value]
        return [add_extension(self, path) for path in value]

    di.FilesOpenItem.from_string = from_string


def patch_choice_widget():
    """callback now returns value aka label, not numerical index of choice"""
    class PatchedChoiceWidget(qtiw.ChoiceWidget):

        def index_changed(self, index):
            if self.store:
                self.store.set(self.item.instance, self.item.item, self.value())
                self.parent_layout.refresh_widgets()
            cb = self.item.get_prop_value("display", "callback", None)
            if cb is not None:
                if self.build_mode:
                    self.set()
                else:
                    self.parent_layout.update_dataitems()
                # the next two lines are the modified ones for pushing the choice value to the
                # callback, not its index:
                choices = self.item.get_prop_value("data", "choices")
                cb(self.item.instance, self.item.item, choices[self.value()][1])
                self.parent_layout.update_widgets(except_this_one=self)

    # this overwrites existing registration:
    qtw.DataSetEditLayout.register(di.ChoiceItem, PatchedChoiceWidget)


def patch_multiplie_choice_widget():
    """implements missing call to callback function"""

    class PatchedMultipleChoiceWidget(qtiw.MultipleChoiceWidget):

        """Adds callback call to MultipleChoiceWidget which is not implemented for this in guidata
        """

        def __init__(self, item, parent_layout):
            super(PatchedMultipleChoiceWidget, self).__init__(item, parent_layout)
            for box in self.boxes:
                QWidget.connect(box, SIGNAL("toggled(bool)"), self.choice_toggled)

        def choice_toggled(self, flag):
            flags = tuple(box.isChecked() for box in self.boxes)
            choices = self.item.get_prop_value("data", "choices")
            value = tuple(c[1] for (c, flag) in zip(choices, flags) if flag)
            cb = self.item.get_prop_value("display", "callback", None)
            if cb is not None:
                if self.build_mode:
                    self.set()
                else:
                    self.parent_layout.update_dataitems()
                cb(self.item.instance, self.item.item, value)
                self.parent_layout.update_widgets(except_this_one=self)

    # this overwrites existing registration:
    qtw.DataSetEditLayout.register(di.MultipleChoiceItem, PatchedMultipleChoiceWidget)


def patch_checkbox_widget():
    """implements missing call to callback function"""

    class PatchedCheckBoxWidget(qtiw.CheckBoxWidget):

        """Adds callback call to BoolItemWidget which is not implemented for this in guidata
        """

        def __init__(self, item, parent_layout):
            super(PatchedCheckBoxWidget, self).__init__(item, parent_layout)
            QWidget.connect(self.checkbox, SIGNAL("toggled(bool)"), self.toggled)

        def toggled(self, value):
            cb = self.item.get_prop_value("display", "callback", None)
            if cb is not None:
                if self.build_mode:
                    self.set()
                else:
                    self.parent_layout.update_dataitems()
                cb(self.item.instance, self.item.item, value)
                self.parent_layout.update_widgets(except_this_one=self)

    # this overwrites existing registration:
    qtw.DataSetEditLayout.register(di.BoolItem, PatchedCheckBoxWidget)


def apply_patches():
    patch_bugs()
    patch_choice_widget()
    patch_multiplie_choice_widget()
    patch_checkbox_widget()



# encoding: utf-8

import collections

import guidata.dataset.dataitems as di
import guidata.dataset.datatypes as dt

from . import _context_stack
from .base import ParameterSet


def begin_group(dct, name, is_expert_group, use_tabs):
    if is_expert_group:
        is_expert_value = dt.ValueProp(False)
        expert_checkbox = di.BoolItem(label="enable expert settings")
        expert_checkbox = expert_checkbox.set_prop("display", store=is_expert_value)
        expert_checkbox._name = "is_expert"
        dct["is_expert_%s" % name] = expert_checkbox
    if use_tabs:
        item = dt.BeginTabGroup(name)
    else:
        item = dt.BeginGroup(name)
    item._name = name
    if is_expert_group:
        item = item.set_prop("display", active=is_expert_value)
    dct["bg_%s" % name] = item


def end_group(dct, name, use_tabs):
    item = dt.EndTabGroup(name) if use_tabs else dt.EndGroup(name)
    item._name = name
    dct["eg_%s" % name] = item


def build_guidata_items(parameter_set, dct, use_tabs=False):
    """
    g is item.group_stack, "bg" means "begin group", "eg" is "end group",
    how gui items are created:

        item(g=())  render_item
        item(g=(1)) bg_1
                    render_item
        item(g=(2)) eg_1
                    bg_2
                    render_item

        item(g=(2, 3)) bg_3
                    render_item

        item(g=(2)) eg_3
                    render_item

        item(g=(2, 4, 5)) bg_4
                        bg_5
                        render_item

        item(g=(2)) eg_5
                    eg_4
                    render_item
        eg_2
    """

    current_stack = []
    for item_name, item in parameter_set._items.items():
        if isinstance(item, ParameterSet):
            continue

        if item._group_stack != tuple(current_stack):
            while len(current_stack) and len(item._group_stack) <= len(current_stack):
                # POP = close groups
                tmp_name = current_stack.pop()
                end_group(dct, tmp_name, use_tabs)
            while len(item._group_stack) > len(current_stack):
                # PUSH = open groups
                tmp_name = item._group_stack[len(current_stack)]
                current_stack.append(tmp_name)
                key = tuple(current_stack)
                is_expert_group = key in _context_stack.current_context().expert_groups
                begin_group(dct, tmp_name, is_expert_group, use_tabs)
        gui_item = item.setup_gui_item()
        gui_item._name = ""
        dct[item_name] = gui_item

    # POP = close groups until stack is empty
    while len(current_stack):
        close = current_stack[-1]
        del current_stack[-1]
        end_group(dct, close, use_tabs)


def create_widget(parameter_set, use_tabs):
    dct = collections.OrderedDict()
    build_guidata_items(parameter_set, dct, use_tabs)
    clz = type("DataSet_%s" % parameter_set.__class__.__name__, (dt.DataSet,), dct)
    clz.__doc__ = parameter_set.__doc__
    return clz


def linearize(parameter_set):

    def _linearize(parameter_set, li):
        li.append(parameter_set)
        for name, item in parameter_set._items.items():
            if isinstance(item, ParameterSet):
                _linearize(item, li)

    li = []
    _linearize(parameter_set, li)
    return li


"""
todo:
  + checks !
  + propagate change
  + title, helpstring, docstring
  + andere items

  - cleanup
  - tests

  - presets
    -> Dialog patchen oder combo box mit callback erzeugen welcher "alles macht" =
  + i/o : via pickle
"""


def create_dlg(c, use_tabs=False):

    dlgs = []

    # checked_propagation: erst nach dem edit fertig ist propagations bestimmen
    # und dialog mit "uebernehmen button" anbieten

    for i, ci in enumerate(linearize(c)):
        dlg = create_widget(ci, use_tabs)(ci._name or "")
        dlgs.append(dlg)
    g = dt.DataSetGroup(dlgs, title='Parameters group')
    return g



import types
import os

import guidata.dataset.dataitems as di

from .base import ConfigItem


class Value(ConfigItem):

    def __init__(self, default=None, **options):
        super(Value, self).__init__(name="")
        self.value = default
        self.target = None
        self.options = options
        if default is not None:
            self._check_default(default)

    def _check_default(self, default):
        if not self.check_type(default):
            raise Exception("can not set %r to value %r because of type mismatch" % (self, value))
        if not self.check_value(default):
            raise Exception("can not set %r to value %r, value check fails" % (self, value))

    def setup_value(self, value):
        return value

    def callback(self, data_set, gui_item, value):
        value_item = gui_item.get_prop("extension", "value_item")
        value_item.value = value
        target = value_item.target
        if target is not None:
            to, method = target
            to.value = method(value)  # sets value to value_item
            setattr(data_set, to._name, method(value))  # sets value to gui_item

    def setup_gui_item(self):
        # we are changing options below, so we work with a copy
        options = self.options.copy()
        if "label" not in options:
            options["label"] = self._name
        options["default"] = self.value
        gui_item = self.create_gui_item(options)
        gui_item.set_prop("display", callback=self.callback)
        return wrap(gui_item, self)

    def __get__(self, from_, type_):
        return self.value


def wrap(gui_item, value_item):
    """set property for retreiving value_item from gui_item
       and pathc gui_items check method:
    """

    # we monkey patch the check_value method but still want to have access to the originial
    # method from guidata as it implements many features which we want to use too, eg checking for
    # a min or max value, etc...
    # binding the original method to a parameter does the trick which was not obvious, so test
    # carefully if you have a reason to change this:

    def check_value(self, value, orig_check_value=gui_item.check_value):
        value_item = self.get_prop("extension", "value_item")
        ok_value_item = hasattr(value_item, "check_value") and value_item.check_value(value)
        ok_gui_item = orig_check_value(value)
        return ok_value_item and ok_gui_item

    gui_item.set_prop("extension", value_item=value_item)
    gui_item.check_value = types.MethodType(check_value, gui_item)
    return gui_item


class NumberValue(Value):

    def check_value(self, value):
        min_ = self.options.get("min")
        max_ = self.options.get("max")
        if min_ is not None and value < min_:
            return False
        if max_ is not None and value > max_:
            return False
        return True


class IntValue(NumberValue):

    def check_type(self, value):
        return isinstance(value, int)

    def create_gui_item(self, options):
        return di.IntItem(**options)


class FloatValue(NumberValue):

    def check_type(self, value):
        return isinstance(value, (int, long, float))

    def create_gui_item(self, options):
        return di.FloatItem(**options)


class SecondsAsMinutes(FloatValue):

    """SecondsAsMinutes: internal value is in seconds, but in GUI we have minutes"""

    def create_gui_item(self, options):
        options = options.copy()
        options["default"] = self.value / 60.0
        min_ = options.get("min")
        if min_ is not None:
            options["min"] = min_ / 60.0
        max_ = options.get("max")
        if max_ is not None:
            options["max"] = max_ / 60.0

        # pass value in seconds to gui item as minutes:
        if "round_digits" in options:
            round_digits = options.pop("round_digits")
            for k in ("default", "min", "max"):
                options[k] = round(options[k], round_digits)

        item = di.FloatItem(**options)
        return item

    def callback(self, data_set, gui_item, value):
        value_item = gui_item.get_prop("extension", "value_item")
        value_item.value = value * 60.0
        target = value_item.target
        if target is not None:
            to, method = target
            to.value = method(value)  # sets value to value_item
            setattr(data_set, to._name, method(value))  # sets value to gui_item


class StringValue(Value):

    def check_type(self, value):
        return isinstance(value, basestring)

    def check_value(self, value):
        return True

    def create_gui_item(self, options):
        return di.StringItem(**options)


class IntListValue(StringValue):

    def check_type(self, value):
        print("check_type", repr(value))
        return isinstance(value, (tuple, list)) and all(isinstance(i, (int, long)) for i in value)

    def check_value(self, value):
        min_len = self.options.get("min_length", 0)
        if isinstance(value, (list, tuple)):
            return len(value) >= min_len and all(isinstance(i, (int, long)) for i in value)
        assert isinstance(value, basestring)
        values = [v.strip() for v in value.split(",")]
        try:
            map(int, values)
        except:
            return False
        else:
            return len(values) >= min_len

    def create_gui_item(self, options):
        options = options.copy()
        if "min_length" in options:
            del options["min_length"]
        options["default"] = ", ".join(map(str, self.value))
        item = di.StringItem(**options)
        return item

    def callback(self, data_set, gui_item, value):
        value_item = gui_item.get_prop("extension", "value_item")
        fields = [v.strip() for v in value.split(",")]
        value_item.value = map(int, fields)
        target = value_item.target
        if target is not None:
            to, method = target
            to.value = method(value)  # sets value to value_item
            setattr(data_set, to._name, method(value))  # sets value to gui_item


class BoolValue(Value):

    def check_type(self, value):
        return isinstance(value, bool)

    def check_value(self, value):
        return True

    def create_gui_item(self, options):
        return di.BoolItem(**options)


class _FiniteDomainValue(Value):

    def __init__(self, choices, *a, **kw):
        # order matters as Value.__init__ introduces check which needs choices attributes
        self.choices = choices
        super(_FiniteDomainValue, self).__init__(*a, **kw)

    def check_type(self, value):
        return True

    def check_value(self, value):
        return value in self.choices


class SingleValued(_FiniteDomainValue):

    def create_gui_item(self, options):
        if not all(isinstance(c, basestring) for c in self.choices):
            raise Exception("MultipleValued only works with strings as choices")
        options = options.copy()
        options["choices"] = self.choices
        if self.value is not None and self.value in self.choices:
            options["default"] = self.choices.index(self.value)
        return di.ChoiceItem(**options)

    def _setup_value(self, value):
        if isinstance(value, int):
            # if value is int and all choices are not int: choose by index:
            if any(isinstance(c, int) for c in self.choices):
                return value
            try:
                new_value = self.choices[value]
            except IndexError:
                raise IndexError("index %d is out of range for %r" % (value, self))
            return new_value
        return super(SingleValued, self).setup_value(value)


class MultipleValued(_FiniteDomainValue):

    def create_gui_item(self, options):
        if not all(isinstance(c, basestring) for c in self.choices):
            raise Exception("MultipleValued only works with strings as choices")
        options = options.copy()
        options["choices"] = self.choices
        cols = options.get("cols")
        rows = options.get("rows")
        if cols:
            del options["cols"]
        if rows:
            del options["rows"]
        if self.value is not None:
            options["default"] = [i for (i, v) in enumerate(self.choices) if v in self.value]
        gui_item = di.MultipleChoiceItem(**options)
        if cols:
            gui_item = gui_item.vertical(cols)
        if rows:
            gui_item = gui_item.horizontal(rows)
        return gui_item

    def _setup_value(self, value):
        if all(isinstance(v, int) for v in value):
            # if value is int and all choices are not int: choose by index:
            if any(isinstance(c, int) for c in self.choices):
                return value
            new_values = []
            try:
                for v in value:
                    new_value = self.choices[v]
                    new_values.append(new_value)
            except IndexError:
                raise IndexError("index %d is out of range for %r" % (value, self))
            return new_values
        return super(MultipleValued, self).setup_value(value)

    def check_type(self, value):
        return True

    def check_value(self, value):
        return all(vi in self.choices for vi in value)


class ExistingFileValue(StringValue):

    def check_type(self, value):
        return isinstance(value, basestring)

    def check_value(self, value):
        return os.path.exists(value)

    def create_gui_item(self, options):
        return di.FileOpenItem(**options)


class ExistingFilesValue(StringValue):

    def check_type(self, value):
        return isinstance(value, (tuple, list)) and all(isinstance(v, basestring) for v in value)

    def check_value(self, value):
        return all(os.path.exists(v) for v in value)

    def create_gui_item(self, options):
        options = options.copy()
        options["default"] = options.get("default") or [""]
        return di.FilesOpenItem(**options)


class FileForSaveValue(StringValue):

    def check_type(self, value):
        return isinstance(value, basestring)

    def check_value(self, value):
        dirname = os.path.dirname(value)
        return os.path.exists(dirname) and os.path.isdir(dirname)

    def create_gui_item(self, options):
        options = options.copy()
        options["default"] = options.get("default") or ""
        return di.FileSaveItem(**options)


class DirectoryValue(StringValue):

    def create_gui_item(self, options):
        options = options.copy()
        options["default"] = options.get("default") or ""
        return di.DirectoryItem(**options)



from functools import wraps

from .base import ParameterSet


def inject_params(fun):
    @wraps(fun)
    def wrapped(*a, **kw):
        names = fun.func_code.co_varnames
        dct = dict()
        psets = list()
        for (name, ai) in zip(names, a):
            if isinstance(ai, ParameterSet):
                dds.append(ai)
            else:
                dct[name] = ai
        for dd in dds:
            dct.update(dd.as_dict())
        dct.update(kw)
        return fun(**dct)
    return wrapped




import contextlib


@contextlib.contextmanager
def ParameterSetBuilder():
    from . import _context_stack

    from .value_items import (IntValue, FloatValue, BoolValue, StringValue, SingleValued,
                              MultipleValued, ExistingFileValue, ExistingFilesValue,
                              FileForSaveValue, DirectoryValue, SecondsAsMinutes,
                              IntListValue)

    from .base import ParameterSet
    from .decl_helpers import propagate_change, ValueGroup

    # create and install new module global context
    _context_stack.install_new_context()

    # import classes and functions
    to_install = (IntValue, FloatValue, BoolValue, StringValue, SingleValued, MultipleValued,
                  ExistingFileValue, ExistingFilesValue, FileForSaveValue, DirectoryValue,
                  ParameterSet, ValueGroup, propagate_change, ValueGroup, SecondsAsMinutes,
                  IntListValue)

    already_there = dict()
    for obj in to_install:
        name = obj.__name__
        if name in __builtins__.keys():
            # remember already global classes and functions before overriding
            already_there[name] = __builtins__.get(name)
        __builtins__[name] = obj
    yield

    # reset overriden global names
    for name, obj in already_there.items():
        __builtins__[name] = obj

    # restore global context
    _context_stack.drop_context()
